import datetime
import typing

import pydantic

from ._base_api import BaseAPIClient


class EvaluateEvaluator(pydantic.BaseModel):
    evaluator: str
    profile_name: str | None = None
    explain_strategy: str = "always"


class EvaluateRequest(pydantic.BaseModel):
    evaluators: list[EvaluateEvaluator]
    evaluated_model_system_prompt: str | None = None
    evaluated_model_retrieved_context: list[str] | None = None
    evaluated_model_input: str | None = None
    evaluated_model_output: str | None = None
    evaluated_model_gold_answer: str | None = None

    app: str
    capture: str = "all"
    tags: dict[str, str] | None = None


class EvaluationResultAdditionalInfo(pydantic.BaseModel):
    positions: list | None
    extra: dict | None
    confidence_interval: dict | None


class EvaluationResult(pydantic.BaseModel):
    id: str
    app: str
    created_at: pydantic.AwareDatetime
    evaluator_id: str
    evaluated_model_system_prompt: str | None
    evaluated_model_retrieved_context: list[str] | None
    evaluated_model_input: str | None
    evaluated_model_output: str | None
    evaluated_model_gold_answer: str | None
    pass_: bool | None = pydantic.Field(alias="pass")
    score_raw: float | None
    additional_info: EvaluationResultAdditionalInfo
    explanation: str | None
    evaluation_duration: datetime.timedelta | None
    explanation_duration: datetime.timedelta | None
    evaluator_family: str
    evaluator_profile_public_id: str
    tags: dict[str, str] | None


class EvaluateResult(pydantic.BaseModel):
    evaluator_id: str
    profile_name: str
    status: str
    error_message: str | None
    evaluation_result: EvaluationResult


class EvaluateResponse(pydantic.BaseModel):
    results: list[EvaluateResult]


class ExportEvaluationResult(pydantic.BaseModel):
    app: str
    evaluator_id: str
    profile_name: str | None = None
    evaluated_model_system_prompt: str | None = None
    evaluated_model_retrieved_context: list[str] | None = None
    evaluated_model_input: str | None = None
    evaluated_model_output: str | None = None
    evaluated_model_gold_answer: str | None = None
    pass_: bool | None = pydantic.Field(alias="pass_", serialization_alias="pass")
    score_raw: float | None
    evaluation_duration: datetime.timedelta | None = None
    evaluated_model_name: str | None = None
    evaluated_model_provider: str | None = None
    evaluated_model_params: dict[str, str | int | float] | None = None
    evaluated_model_selected_model: str | None = None
    tags: dict[str, str] | None = None


class ExportEvaluationRequest(pydantic.BaseModel):
    evaluation_results: list[ExportEvaluationResult]


class ExportEvaluationResultPartial(pydantic.BaseModel):
    id: str
    app: str
    created_at: pydantic.AwareDatetime
    evaluator_id: str


class ExportEvaluationResponse(pydantic.BaseModel):
    evaluation_results: list[ExportEvaluationResultPartial]


class ListProfilesRequest(pydantic.BaseModel):
    public_id: str | None = None
    evaluator_family: str | None = None
    evaluator_id: str | None = None
    name: str | None = None
    revision: str | None = None
    get_last_revision: bool = False
    is_patronus_managed: bool | None = None
    limit: int = 1000
    offset: int = 0


class EvaluatorProfile(pydantic.BaseModel):
    public_id: str
    evaluator_family: str
    name: str
    revision: int
    config: dict[str, typing.Any] | None
    is_patronus_managed: bool
    created_at: datetime.datetime
    description: str | None


class ListProfilesResponse(pydantic.BaseModel):
    evaluator_profiles: list[EvaluatorProfile]


class DatasetDatum(pydantic.BaseModel):
    dataset_id: str
    sid: int
    evaluated_model_system_prompt: str | None = None
    evaluated_model_retrieved_context: list[str] | None = None
    evaluated_model_input: str | None = None
    evaluated_model_output: str | None = None
    evaluated_model_gold_answer: str | None = None
    meta_evaluated_model_name: str | None = None
    meta_evaluated_model_provider: str | None = None
    meta_evaluated_model_selected_model: str | None = None
    meta_evaluated_model_params: dict[str, str | int | float] | None = None


class ListDatasetData(pydantic.BaseModel):
    data: list[DatasetDatum]


class API(BaseAPIClient):
    async def evaluate(self, request: EvaluateRequest) -> EvaluateResponse:
        resp = await self.call("POST", "/v1/evaluate", body=request, response_cls=EvaluateResponse)
        # TODO error handling
        resp.response.raise_for_status()
        return resp.data

    async def export_evaluations(self, request: ExportEvaluationRequest) -> ExportEvaluationResponse:
        resp = await self.call(
            "POST",
            "/v1/evaluation-results/batch",
            body=request,
            response_cls=ExportEvaluationResponse,
        )
        # TODO error handling
        resp.response.raise_for_status()
        return resp.data

    async def get_profile(self, evaluator_family: str, name: str) -> EvaluatorProfile:
        profiles = await self.list_profiles(
            ListProfilesRequest(
                evaluator_family=evaluator_family,
                name=name,
                get_last_revision=True,
            )
        )
        # TODO error handling
        assert (
            len(profiles.evaluator_profiles) == 1
        ), f"get_profile didn't return 1 profile. It returned {len(profiles.evaluator_profiles)!r} instead"
        return profiles.evaluator_profiles[0]

    async def list_profiles(self, request: ListProfilesRequest) -> ListProfilesResponse:
        params = request.model_dump(exclude_none=True)
        resp = await self.call(
            "GET",
            "/v1/evaluator-profiles",
            params=params,
            response_cls=ListProfilesResponse,
        )
        # TODO error handling
        resp.response.raise_for_status()
        return resp.data

    async def list_dataset_data(self, dataset_id: str):
        resp = await self.call("GET", f"/v1/datasets/{dataset_id}/data", response_cls=ListDatasetData)
        # TODO error handling
        resp.response.raise_for_status()
        return resp.data



import asyncio
import functools
import inspect
import typing
from concurrent.futures import ThreadPoolExecutor

T = typing.TypeVar("T")


async def run_as_coro(
    __loop: asyncio.AbstractEventLoop,
    __executor: ThreadPoolExecutor,
    __fn: typing.Callable[[typing.Any, ...], T],
    *args,
    **kwargs,
) -> typing.Awaitable[T]:
    if inspect.iscoroutinefunction(__fn):
        return await __fn(*args, **kwargs)

    @functools.wraps(__fn)
    def inner(args: tuple[typing.Any, ...], kwargs: dict[str, typing.Any]):
        return __fn(*args, **kwargs)

    return await __loop.run_in_executor(__executor, inner, args, kwargs)


def run_until_complete(coro):
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.get_event_loop()

    return loop.run_until_complete(coro)



import asyncio
import dataclasses
import inspect
import abc
import time
import typing
from concurrent.futures import ThreadPoolExecutor

from ._async_utils import run_as_coro

EVALUATION_ARGS = [
    "app",
    "evaluated_model_system_prompt",
    "evaluated_model_retrieved_context",
    "evaluated_model_input",
    "evaluated_model_output",
    "evaluated_model_gold_answer",
    "tags",
]

EvalF = typing.TypeVar("EvalF", bound=typing.Callable[..., typing.Any])


class EvaluationResultT(typing.Protocol):
    pass_: bool | None
    score_raw: float | None
    tags: dict[str, str] | None = None


@dataclasses.dataclass
class EvaluationResult:
    pass_: bool | None
    score_raw: float | None
    tags: dict[str, str] | None = None


@dataclasses.dataclass
class EvaluatorOutput:
    result: EvaluationResultT
    duration: float


# TODO we should have more classes of evaluator errors
#      some of them may be recoverable on retry e.g. on remote 500, 502, 429
#      and some not, e.g. ZeroDivisionError locally, or 400, 422 remote
class EvaluatorError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class Evaluator(abc.ABC):
    name = "unknown"
    profile_name = None
    accepted_args: set[str]
    remote_capture = False

    def __init__(self, accepted_args: set[str] | None = None):
        self.accepted_args = accepted_args
        if not self.accepted_args:
            sig = inspect.signature(self.evaluate)
            param_keys = sig.parameters.keys()
            for name in param_keys:
                if name not in EVALUATION_ARGS:
                    raise ValueError(
                        f"{name!r} is not a valid evaluator argument. Valid arguments are: {EVALUATION_ARGS}"
                    )
            self.accepted_args = set(param_keys)

    def display_name(self) -> str:
        return self.name

    async def execute(
        self,
        loop: asyncio.AbstractEventLoop,
        executor: ThreadPoolExecutor,
        app: str | None = None,
        evaluated_model_system_prompt: str | None = None,
        evaluated_model_retrieved_context: list[str] | None = None,
        evaluated_model_input: str | None = None,
        evaluated_model_output: str | None = None,
        evaluated_model_gold_answer: str | None = None,
        tags: dict[str, str] | None = None,
    ) -> EvaluatorOutput:
        kwargs = {
            "app": app,
            "evaluated_model_system_prompt": evaluated_model_system_prompt,
            "evaluated_model_retrieved_context": evaluated_model_retrieved_context,
            "evaluated_model_input": evaluated_model_input,
            "evaluated_model_output": evaluated_model_output,
            "evaluated_model_gold_answer": evaluated_model_gold_answer,
            "tags": tags,
        }
        pass_kwargs = {k: v for k, v in kwargs.items() if k in self.accepted_args}
        start_time = time.perf_counter()
        result = await run_as_coro(
            __loop=loop,
            __executor=executor,
            __fn=self.evaluate,
            **pass_kwargs,
        )
        end_time = time.perf_counter()
        elapsed = end_time - start_time
        return EvaluatorOutput(result=result, duration=elapsed)

    @abc.abstractmethod
    def evaluate(self, **kwargs) -> EvaluationResultT | typing.Awaitable[EvaluationResultT]: ...


class FunctionalEvaluator(Evaluator):
    fn: EvalF

    def __init__(self, name: str, fn: EvalF, accepted_args: set[str]):
        self.name = name
        self.fn = fn
        super().__init__(accepted_args)

    def evaluate(self, **kwargs) -> EvaluationResultT:
        ret = self.fn(**kwargs)
        if isinstance(ret, bool):
            return EvaluationResult(pass_=ret, score_raw=float(ret))
        return ret


def evaluator(fn: EvalF, evaluator_name: str | None = None) -> FunctionalEvaluator:
    sig = inspect.signature(fn)
    param_keys = sig.parameters.keys()
    for name in param_keys:
        if name not in EVALUATION_ARGS:
            raise ValueError(f"{name!r} is not a valid evaluator argument. Valid arguments are: {EVALUATION_ARGS}")
    evaluator_name = evaluator_name or fn.__name__
    return FunctionalEvaluator(evaluator_name, fn, set(param_keys))


def simple_evaluator(fn: typing.Callable[[str, str], bool], name: str | None = None) -> FunctionalEvaluator:
    """
    Simple evaluator allows to wrap functions boolean evaluation:

    ```python
    exact_match = simple_evaluator(lambda output, gold_answer: output == gold_answer)
    ```

    It can be also used to decorate multi-line function.

    ```python
    @simple_evaluator
    def iexact_match(output, gold_answer) -> bool:
        output = output.strip().lower()
        gold_answer = gold_answer.strip().lower()
        return output == gold_answer
    ```
    """

    def wrapper(evaluated_model_output: str, evaluated_model_gold_answer: str) -> EvaluationResultT:
        passed = fn(evaluated_model_output, evaluated_model_gold_answer)
        return EvaluationResult(pass_=passed, score_raw=float(passed))

    name = name or fn.__name__
    if name == "<lambda>":
        name = "simple_evaluator"
    return evaluator(wrapper, evaluator_name=name)



__all__ = [
    "Client",
    "Task",
    "task",
    "simple_task",
    "evaluator",
    "simple_evaluator",
    "Evaluator",
    "EvaluationResult",
    "TaskResult",
]

from ._evaluators import evaluator, Evaluator, EvaluationResult, simple_evaluator
from ._tasks import Task, task, simple_task, TaskResult
from ._client import Client



import os
import typing

import httpx

from ._async_utils import run_until_complete
from ._evaluators import Evaluator
from ._evaluators_remote import RemoteEvaluator
from ._tasks import Task
from ._api import API


class Client:
    def __init__(
        self,
        api_key: str | None = None,
        base_url: str = "https://api.patronus.ai",
        api: API | None = None,
        # TODO Allow passing more types for the timeout: float, Timeout, None, NotSet
        timeout: float = 300,
    ):
        if api_key is None:
            api_key = os.environ.get("PATRONUSAI_API_KEY")

        if not api_key:
            raise ValueError("Provide 'api_key' argument or set PATRONUSAI_API_KEY environment variable.")

        if api is None:
            # TODO allow passing http client as an argument
            http_client = httpx.AsyncClient(timeout=timeout)

            # TODO use package version
            api = API(version="0.0.1", http=http_client)
        api.set_target(base_url, api_key)
        self.api = api

    def experiment(
        self,
        name: str,
        data: list[dict],
        task: Task,
        evaluators: list[Evaluator],
        tags: dict[str, str] | None = None,
        display_hist: bool = False,
    ):
        from ._experiment import experiment as ex

        ex(
            self,
            name=name,
            data=data,
            task=task,
            evaluators=evaluators,
            tags=tags,
            display_hist=display_hist,
        )

    def remote_evaluator(
        self,
        # TODO family should not be necessary but we need to get aliases from the API
        family: str,
        # ID or an alias of an evaluator.
        evaluator: str,
        profile_name: str,
        # TODO
        # config: dict[str, typing.Any] | None = None,
        # allow_upsert: bool = False,
    ) -> RemoteEvaluator:
        # TODO this should be awaited
        profile = run_until_complete(self.api.get_profile(family, profile_name))
        return RemoteEvaluator(
            evaluator=evaluator,
            profile_name=profile.name,
            api_=self.api,
        )

    def remote_dataset(self, dataset_id: str) -> list[dict[str, typing.Any]]:
        data = run_until_complete(self.api.list_dataset_data(dataset_id))
        return data.model_dump()["data"]

    def get_project_by_name(self, project_name: str):
        # TODO
        pass
